package xyz.matj1.bp;

import mcdr.sequence.impl.SequenceMotionWordsDTW;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtils;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import xyz.matj1.bp.Utils.Pair;

import java.io.File;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Predicate;
import java.util.regex.PatternSyntaxException;
import java.util.stream.IntStream;

/**
 * A class for testing of how motion-word sequences
 * can be matched on artificially stretched sequences
 *
 * @author Matěj Bagar
 */
public class TestingOnStretchedSequences {
	public static void main(String[] args) throws IOException {
		var baseSequenceMgmt = BPBagar2.loadData(
			"data/hdm05-annotations_specific-segment80_shift16-coords_normPOS-fps12-quantized-pivots-kmedoids-350.data",
			SequenceMotionWordsDTW.class
		);
		var stretchedSequenceMgmt = BPBagar2.loadData(
			"data/hard_motion_words_stretched_2×.data",
			SequenceMotionWordsDTW.class
		);
		var stretchMultiple = 2;
		var outputDir = new File("output/graphs for stretched sequences/all.png");

		stretchMatchingShow(baseSequenceMgmt, stretchedSequenceMgmt, stretchMultiple, outputDir);
	}

	/**
	 * TODO
	 * @param baseSequenceMgmt
	 * @param stretchedSequenceMgmt
	 * @param stretchMultiple
	 * @param outputFile
	 * @throws IOException
	 */
	public static void stretchMatchingShow(
		ObjectMgmt baseSequenceMgmt,
		ObjectMgmt stretchedSequenceMgmt,
		double stretchMultiple,
		File outputFile
	) throws IOException {
		var results = stretchMatching(baseSequenceMgmt, stretchedSequenceMgmt);
		var dataSeries = new ConcurrentHashMap<Category, XYSeries>();

		results.entrySet().parallelStream().forEach(entry -> {
			var length = Integer.parseInt(entry.getKey().split("_")[3]);
			var category = entry.getValue().a();
			var distance = entry.getValue().b();

			if (!dataSeries.containsKey(category)) {
				dataSeries.put(category, new XYSeries(category.index));
			}
			var series = dataSeries.get(category);
			series.add(length, distance);
		});

		var dataset = new XYSeriesCollection();
		dataSeries.values().forEach(dataset::addSeries);

		var chart = ChartFactory.createScatterPlot(
			"The length of a sequence and the distance to its " + stretchMultiple + "× stretched version",
			"the length",
			"the distance",
			dataset,
			PlotOrientation.VERTICAL,
			true,
			false,
			false
		);

		ChartUtils.saveChartAsPNG(outputFile, chart, 2048, 1024);

		System.err.println("Stretch-matching graph written to " + outputFile.getPath());
	}

	/**
	 * This returns a {@link Map} of its locator URI and a
	 * {@link Pair} of its category and the distance to its stretched version
	 * for all sequences in <code>baseSequenceMgmt</code>.
	 * <code>stretchedSequenceMgmt</code> is supposed to contain the stretched versions
	 * of all sequences in <code>baseSequenceMgmt</code> in the same categories.
	 *
	 * @param baseSequenceMgmt a management of the base sequences
	 * @param stretchedSequenceMgmt a management of the stretched vesions of the sequences
	 * @return a {@link Map} of its locator URI and a {@link Pair} of its category and the distance to its stretched version for all sequences in <code>baseSequenceMgmt</code>
	 */
	public static Map<String, Pair<Category, Double>> stretchMatching(ObjectMgmt baseSequenceMgmt, ObjectMgmt stretchedSequenceMgmt) {
		var results = new ConcurrentHashMap<String, Pair<Category, Double>>();
		var stretchedCategoryObjects = stretchedSequenceMgmt.getCategoryObjects();

		for (var entry : baseSequenceMgmt.getCategoryObjects().entrySet()) {
			var category = entry.getKey();
			var sequences = entry.getValue();
			for (var sequence : sequences) {
				var stretchedSequence = stretchedCategoryObjects.get(category).stream()
					.filter(uriMatches(sequence)).findFirst().get();
				results.put(
					sequence.getLocatorURI(),
					new Pair<>(category, (double)sequence.getDistance(stretchedSequence))
				);
		}}

		return results;
	}

	/**
	 * This returns a {@link Predicate} that tells if its given {@link LocalAbstractObject}
	 * has a locator URI matching to the locator URI of <code>sequence</code>.
	 * Two LocalAbstractObjects have matching locator URIs iff
	 * the URIs are digit sequences separated by “_” and the a three of them are equal.
	 * This is supposed to be used to search for matching stretched sequences.
	 *
	 * @param sequence a sequence to whose URI the other's URI shall be matched
	 * @return a predicate that tells if its given object has a locator URI matching to that of <code>sequence</code>
	 */
	public static Predicate<LocalAbstractObject> uriMatches(LocalAbstractObject sequence) {
		return other -> {
			try {
				var uriArray = sequence.getLocatorURI().split("_");
				var otherUriArray = other.getLocatorURI().split("_");

				return IntStream.of(0, 1, 2).allMatch(
					x -> uriArray[x].equals(otherUriArray[x])
				);
			} catch (PatternSyntaxException | NumberFormatException | IndexOutOfBoundsException e) {
				return false;
	}}; }
}
